{% extends "common/mailheaders.html" %}{% load i18n %}
{% block extraheaders %}{% if attachments %}
  <div class="row">
    <div class="header">{% trans "Attachments" %}</div>
    <div class="value">{% url "modoboa_webmail:attachment_get" as url %}
      {% for key, fname in attachments.items %}
      <a href="{{ url }}?mbox={{ folder }}&mailid={{ mail_id }}&fname={{ fname }}&partnumber={{ key }}">{{ fname }}</a>
      {% endfor %}
    </div>
  </div>
{% endif %}{% endblock %}
